import inspect

from trulens.core.utils.imports import Dummy
from trulens.core.utils.imports import OptionalImports
from trulens.core.utils.imports import format_import_errors

with OptionalImports(
    messages=format_import_errors(
        ["ipython", "ipywidgets"], purpose="using TruLens in a notebook"
    )
):
    from IPython import get_ipython
    from IPython.core.magic import register_line_cell_magic
    from IPython.display import display
    from ipywidgets import widgets


def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except Exception:
        return False


def setup_widget_stdout_stderr():
    out_stdout = widgets.Output()
    out_stderr = widgets.Output()

    acc = widgets.Accordion(
        children=[
            widgets.VBox([
                widgets.VBox([widgets.Label("STDOUT"), out_stdout]),
                widgets.VBox([widgets.Label("STDERR"), out_stderr]),
            ])
        ],
        open=True,
    )
    acc.set_title(0, "Dashboard log")

    display(acc)
    return out_stdout, out_stderr


if not isinstance(register_line_cell_magic, Dummy) and is_notebook():

    @register_line_cell_magic
    def writefileinterpolated(line, cell):
        caller_frame = inspect.stack()[2]
        caller_globals = caller_frame.frame.f_globals

        with open(line, "w") as f:
            f.write(cell.format(**caller_globals))



from functools import partial
import hashlib
import pprint as pp
from typing import Any, Dict, List, Optional, Sequence

import pandas as pd
import streamlit as st
from streamlit_pills import pills
from trulens.core.database.base import MULTI_CALL_NAME_DELIMITER
from trulens.dashboard.display import expand_groundedness_df
from trulens.dashboard.display import highlight
from trulens.dashboard.ux.styles import CATEGORY
from trulens.dashboard.ux.styles import default_direction


def df_cell_highlight(
    score: float,
    feedback_name: str,
    feedback_directions: Dict[str, bool],
    n_cells: int = 1,
):
    if "distance" in feedback_name:
        return [f"background-color: {CATEGORY.UNKNOWN.color}"] * n_cells
    cat = CATEGORY.of_score(
        score,
        higher_is_better=feedback_directions.get(
            feedback_name, default_direction == "HIGHER_IS_BETTER"
        ),
    )
    return [f"background-color: {cat.color}"] * n_cells


def display_feedback_call(
    record_id: str,
    call: List[Dict[str, Any]],
    feedback_name: str,
    feedback_directions: Dict[str, bool],
):
    if call is not None and len(call) > 0:
        # NOTE(piotrm for garett): converting feedback
        # function inputs to strings here as other
        # structures get rendered as [object Object] in the
        # javascript downstream. If the first input/column
        # is a list, the DataFrame.from_records does create
        # multiple rows, one for each element, but if the
        # second or other column is a list, it will not do
        # this.
        for c in call:
            args: Dict = c["args"]
            for k, v in args.items():
                if not isinstance(v, str):
                    args[k] = pp.pformat(v)

        df = pd.DataFrame.from_records(c["args"] for c in call)

        df["score"] = pd.DataFrame([
            float(call[i]["ret"]) if call[i]["ret"] is not None else -1
            for i in range(len(call))
        ])
        df["meta"] = pd.Series([call[i]["meta"] for i in range(len(call))])
        df = df.join(df.meta.apply(lambda m: pd.Series(m))).drop(columns="meta")

        # note: improve conditional to not rely on the feedback name
        if "groundedness" in feedback_name.lower():
            df = expand_groundedness_df(df)
        style_highlight_fn = partial(
            highlight,
            selected_feedback=feedback_name,
            feedback_directions=feedback_directions,
            default_direction=default_direction,
        )
        styled_df = df.style.apply(
            style_highlight_fn,
            axis=1,
        )

        # Format only numeric columns
        for col in df.select_dtypes(include=["number"]).columns:
            styled_df = styled_df.format({col: "{:.2f}"})

        st.dataframe(styled_df, hide_index=True)
    else:
        st.warning("No feedback details found.")


def _render_feedback_pills(
    feedback_col_names: Sequence[str],
    feedback_directions: Dict[str, bool],
    selected_row: Optional[pd.Series] = None,
    key_prefix: str = "",
):
    if len(feedback_col_names) == 0:
        st.warning("No feedback details found.")
        return

    if selected_row is not None:

        def get_icon(feedback_name: str):
            cat = CATEGORY.of_score(
                selected_row[feedback_name],
                higher_is_better=feedback_directions.get(feedback_name, True),
            )
            return cat.icon

        feedback_with_valid_results = sorted([
            fcol
            for fcol in feedback_col_names
            if fcol in selected_row and selected_row[fcol] is not None
        ])

        icons = [get_icon(fcol) for fcol in feedback_with_valid_results]
    else:
        feedback_with_valid_results = feedback_col_names

    if len(feedback_with_valid_results) == 0:
        st.warning("No feedback functions found.")
        return

    hash = hashlib.md5(str(feedback_with_valid_results).encode()).hexdigest()
    if selected_row is None:
        return pills(
            "Feedback Functions (click to learn more)",
            feedback_with_valid_results,
            index=None,
            format_func=lambda fcol: f"{fcol}",
            key=f"{key_prefix}_pills_{hash}",
        )

    return pills(
        "Feedback Functions (click to learn more)",
        feedback_with_valid_results,
        index=None,
        format_func=lambda fcol: f"{fcol} {selected_row[fcol]:.4f}",
        icons=icons,
        key=f"{key_prefix}_pills_{selected_row['app_id']}_{hash}",
    )


def _render_feedback_call(
    feedback_col: str,
    selected_row: pd.Series,
    feedback_directions: Dict[str, bool],
):
    fcol = feedback_col
    if MULTI_CALL_NAME_DELIMITER in feedback_col:
        fcol = feedback_col.split(MULTI_CALL_NAME_DELIMITER)[0]

    feedback_calls = selected_row[f"{feedback_col}_calls"]
    display_feedback_call(
        selected_row["record_id"],
        feedback_calls,
        fcol,
        feedback_directions=feedback_directions,
    )



# Cross-page constants to import from here to avoid circular imports

LEADERBOARD_PAGE_NAME = "Leaderboard"
RECORDS_PAGE_NAME = "Records"
COMPARE_PAGE_NAME = "Compare"

RECORDS_LIMIT = 1000

CACHE_TTL = 60 * 15  # 15 minutes

PINNED_COL_NAME = "trulens.dashboard.pinned"
EXTERNAL_APP_COL_NAME = "trulens.dashboard.external_app"
HIDE_RECORD_COL_NAME = "trulens.dashboard.hide_record"



import json
from typing import Dict, List, Optional, Sequence

import pandas as pd
from st_aggrid import AgGrid
from st_aggrid.grid_options_builder import GridOptionsBuilder
from st_aggrid.shared import ColumnsAutoSizeMode
from st_aggrid.shared import DataReturnMode
import streamlit as st
from trulens.dashboard.components.record_viewer import record_viewer
from trulens.dashboard.constants import EXTERNAL_APP_COL_NAME
from trulens.dashboard.constants import HIDE_RECORD_COL_NAME
from trulens.dashboard.constants import PINNED_COL_NAME
from trulens.dashboard.constants import RECORDS_PAGE_NAME as page_name
from trulens.dashboard.utils.dashboard_utils import ST_RECORDS_LIMIT
from trulens.dashboard.utils.dashboard_utils import get_feedback_defs
from trulens.dashboard.utils.dashboard_utils import get_records_and_feedback
from trulens.dashboard.utils.dashboard_utils import (
    read_query_params_into_session_state,
)
from trulens.dashboard.utils.dashboard_utils import render_app_version_filters
from trulens.dashboard.utils.dashboard_utils import render_sidebar
from trulens.dashboard.utils.dashboard_utils import set_page_config
from trulens.dashboard.utils.records_utils import _render_feedback_call
from trulens.dashboard.utils.records_utils import _render_feedback_pills
from trulens.dashboard.ux.styles import aggrid_css
from trulens.dashboard.ux.styles import cell_rules
from trulens.dashboard.ux.styles import default_direction
from trulens.dashboard.ux.styles import radio_button_css


def init_page_state():
    if st.session_state.get(f"{page_name}.initialized", False):
        return

    read_query_params_into_session_state(
        page_name=page_name,
        transforms={
            "app_ids": lambda x: x.split(","),
        },
    )

    app_ids: Optional[List[str]] = st.session_state.get(
        f"{page_name}.app_ids", None
    )
    if app_ids and not st.query_params.get("app_ids", None):
        st.query_params["app_ids"] = ",".join(app_ids)

    st.session_state[f"{page_name}.initialized"] = True


def _format_cost(cost: float, currency: str) -> str:
    if currency == "USD":
        return f"${cost:.2f}"
    else:
        return f"{cost:.3g} {currency}"


def _render_record_metrics(
    records_df: pd.DataFrame, selected_row: pd.Series
) -> None:
    """Render record level metrics (e.g. total tokens, cost, latency) compared
    to the average when appropriate."""

    app_specific_df = records_df[records_df["app_id"] == selected_row["app_id"]]

    token_col, cost_col, latency_col, _ = st.columns([1, 1, 1, 3])

    num_tokens = selected_row["total_tokens"]
    with token_col.container(height=128, border=True):
        st.metric(
            label="Total tokens (#)",
            value=num_tokens,
            help="Number of tokens generated for this record.",
        )

    cost = selected_row["total_cost"]
    cost_currency = selected_row["cost_currency"]
    average_cost = app_specific_df[
        app_specific_df["cost_currency"] == cost_currency
    ]["total_cost"].mean()
    delta_cost = cost - average_cost

    with cost_col.container(height=128, border=True):
        st.metric(
            label=f"Total cost ({cost_currency})",
            value=_format_cost(cost, cost_currency),
            delta=f"{delta_cost:3g}" if delta_cost != 0 else None,
            delta_color="inverse",
            help=f"Cost of the app execution measured in {cost_currency}. Delta is relative to average cost for the app.",
        )

    latency = selected_row["latency"]
    average_latency = app_specific_df["latency"].mean()
    delta_latency = latency - average_latency
    with latency_col.container(height=128, border=True):
        st.metric(
            label="Latency (s)",
            value=f"{selected_row['latency']}s",
            delta=f"{delta_latency:.3g}s" if delta_latency != 0 else None,
            delta_color="inverse",
            help="Latency of the app execution. Delta is relative to average latency for the app.",
        )


@st.fragment
def _render_trace(
    selected_row: pd.Series,
    records_df: pd.DataFrame,
    feedback_col_names: Sequence[str],
    feedback_directions: Dict[str, bool],
):
    # Start the record specific section
    st.divider()

    # Breadcrumbs
    st.caption(f"{selected_row['app_id']} / {selected_row['record_id']}")
    st.markdown(f"#### {selected_row['record_id']}")

    input_col, output_col = st.columns(2)
    with input_col.expander("Record Input"):
        st.code(selected_row["input"], wrap_lines=True)

    with output_col.expander("Record Output"):
        st.code(selected_row["output"], wrap_lines=True)

    _render_record_metrics(records_df, selected_row)

    app_json = json.loads(selected_row["app_json"])
    record_json = json.loads(selected_row["record_json"])

    feedback_results = st.container()
    trace_details = st.container()

    # Feedback results
    feedback_results.subheader("Feedback Results")
    with feedback_results:
        if selected_ff := _render_feedback_pills(
            feedback_col_names=feedback_col_names,
            selected_row=selected_row,
            feedback_directions=feedback_directions,
        ):
            _render_feedback_call(
                selected_ff,
                selected_row,
                feedback_directions=feedback_directions,
            )

    # Trace details

    with trace_details:
        st.subheader("Trace Details")
        record_viewer(record_json, app_json)


def _preprocess_df(
    records_df: pd.DataFrame,
    record_query: Optional[str] = None,
):
    if HIDE_RECORD_COL_NAME in records_df.columns:
        records_df = records_df[~records_df[HIDE_RECORD_COL_NAME]]
    records_df = records_df.sort_values(by="ts", ascending=False)
    records_df["input"] = (
        records_df["input"].str.encode("utf-8").str.decode("unicode-escape")
    )
    records_df["output"] = (
        records_df["output"].str.encode("utf-8").str.decode("unicode-escape")
    )

    # Reorder App columns to the front. Used for aggrid display
    records_df.insert(0, "app_id", records_df.pop("app_id"))
    records_df.insert(0, "app_name", records_df.pop("app_name"))
    records_df.insert(0, "record_id", records_df.pop("record_id"))
    records_df.insert(0, "app_version", records_df.pop("app_version"))

    # if "app_json" in records_df.columns:
    #     records_df = records_df.drop(columns="app_json")

    if record_query:
        records_df = records_df[
            records_df["app_version"].str.contains(record_query, case=False)
            | records_df["input"].str.contains(record_query, case=False)
            | records_df["output"].str.contains(record_query, case=False)
        ]
    records_df["ts"] = pd.to_datetime(records_df["ts"])
    return records_df


def _build_grid_options(
    df: pd.DataFrame,
    feedback_col_names: Sequence[str],
    feedback_directions: Dict[str, bool],
    version_metadata_col_names: Sequence[str],
):
    gb = GridOptionsBuilder.from_dataframe(df, headerHeight=50)

    gb.configure_column(
        "app_version",
        header_name="App Version",
        resizable=True,
        pinned="left",
        flex=3,
        filter="agMultiColumnFilter",
    )
    gb.configure_column(
        "app_name",
        header_name="App Name",
        hide=True,
        filter="agMultiColumnFilter",
    )

    gb.configure_column(
        PINNED_COL_NAME,
        header_name="Pinned",
        hide=True,
        filter="agSetColumnFilter",
    )
    gb.configure_column(
        EXTERNAL_APP_COL_NAME,
        header_name="External App",
        hide=True,
        filter="agSetColumnFilter",
    )

    gb.configure_column(
        "app_id", header_name="App ID", hide=True, filter="agSetColumnFilter"
    )
    gb.configure_column(
        "record_id",
        header_name="Record ID",
        pinned="left",
        hide=True,
        filter="agSetColumnFilter",
    )

    gb.configure_column(
        "input",
        header_name="User Input",
        wrapText=True,
        autoHeight=True,
        filter="agMultiColumnFilter",
    )
    gb.configure_column(
        "output",
        header_name="Response",
        wrapText=True,
        autoHeight=True,
        filter="agMultiColumnFilter",
    )
    gb.configure_column(
        "record_metadata",
        header_name="Record Metadata",
        filter="agTextColumnFilter",
    )

    gb.configure_column(
        "total_tokens",
        header_name="Total Tokens (#)",
        filter="agNumberColumnFilter",
    )
    gb.configure_column(
        "total_cost",
        header_name="Total Cost (App)",
        filter="agNumberColumnFilter",
    )
    gb.configure_column(
        "cost_currency",
        header_name="Cost Currency",
        filter="agMultiColumnFilter",
    )
    gb.configure_column(
        "latency",
        header_name="Latency (Seconds)",
        filter="agNumberColumnFilter",
    )
    gb.configure_column(
        "tags", header_name="Application Tag", filter="agSetColumnFilter"
    )
    gb.configure_column(
        "ts", header_name="Time Stamp", sort="desc", filter="agDateColumnFilter"
    )

    gb.configure_column(
        "feedback_id",
        header_name="Feedback ID",
        hide=True,
        filter="agMultiColumnFilter",
    )
    gb.configure_column(
        "type", header_name="App Type", hide=True, filter="agMultiColumnFilter"
    )
    gb.configure_column(
        "record_json",
        header_name="Record JSON",
        hide=True,
        filter="agTextColumnFilter",
    )
    gb.configure_column(
        "app_json",
        header_name="App JSON",
        hide=True,
        filter="agTextColumnFilter",
    )
    gb.configure_column(
        "cost_json",
        header_name="Cost JSON",
        hide=True,
        filter="agTextColumnFilter",
    )
    gb.configure_column(
        "perf_json",
        header_name="Perf. JSON",
        hide=True,
        filter="agTextColumnFilter",
    )

    for metadata_col in version_metadata_col_names:
        gb.configure_column(
            metadata_col,
            header_name=metadata_col,
            resizable=True,
            pinned=True,
            hide=True,
        )

    for feedback_col in feedback_col_names:
        if "distance" in feedback_col:
            gb.configure_column(feedback_col, hide=False)
            gb.configure_column(feedback_col + "_calls", hide=True)
        else:
            feedback_direction = (
                "HIGHER_IS_BETTER"
                if feedback_directions.get(feedback_col, default_direction)
                else "LOWER_IS_BETTER"
            )
            gb.configure_column(
                feedback_col,
                cellClassRules=cell_rules[feedback_direction],
                hide=False,
            )
            gb.configure_column(
                feedback_col + "_calls",
                hide=True,
            )

    for col in df.columns:
        if "feedback cost" in col:
            gb.configure_column(
                col,
                hide=True,
            )
    gb.configure_grid_options(
        rowHeight=40,
        suppressContextMenu=True,
    )
    gb.configure_selection(
        selection_mode="single",
        use_checkbox=True,
    )
    gb.configure_pagination(enabled=True)
    gb.configure_side_bar()
    return gb.build()


def _render_grid(
    df: pd.DataFrame,
    feedback_col_names: Sequence[str],
    feedback_directions: Dict[str, bool],
    version_metadata_col_names: Sequence[str],
):
    height = 1000 if len(df) > 20 else 45 * len(df) + 100

    return AgGrid(
        df,
        # key="records_data",
        height=height,
        gridOptions=_build_grid_options(
            df=df,
            feedback_col_names=feedback_col_names,
            feedback_directions=feedback_directions,
            version_metadata_col_names=version_metadata_col_names,
        ),
        update_on=["selectionChanged"],
        custom_css={**aggrid_css, **radio_button_css},
        columns_auto_size_mode=ColumnsAutoSizeMode.FIT_CONTENTS,
        data_return_mode=DataReturnMode.FILTERED,
        allow_unsafe_jscode=True,
    )


def _render_grid_tab(
    df: pd.DataFrame,
    feedback_col_names: List[str],
    feedback_directions: Dict[str, bool],
    version_metadata_col_names: List[str],
):
    grid_data = _render_grid(
        df,
        feedback_col_names=feedback_col_names,
        feedback_directions=feedback_directions,
        version_metadata_col_names=version_metadata_col_names,
    )
    selected_rows = grid_data.selected_rows
    selected_records = pd.DataFrame(selected_rows)
    if selected_records.empty:
        selected_record_id = st.session_state.get(
            f"{page_name}.selected_record", None
        )
        selected_records = df[df["record_id"] == selected_record_id]

    if selected_records.empty:
        st.info(
            "Click a record's checkbox to view details.",
            icon="ℹ️",
        )
        return

    selected_record = selected_records.iloc[0]
    st.session_state[f"{page_name}.selected_record"] = selected_record[
        "record_id"
    ]
    st.query_params["selected_record"] = selected_record["record_id"]
    _render_trace(selected_record, df, feedback_col_names, feedback_directions)


def _reset_app_ids():
    del st.session_state[f"{page_name}.app_ids"]
    st.query_params.pop("app_ids")


def _render_app_id_args_filter(versions_df: pd.DataFrame):
    if app_ids := st.session_state.get(f"{page_name}.app_ids", None):
        # Reading session state from other pages
        valid_app_ids = versions_df["app_id"].unique()
        if not set(app_ids).issubset(valid_app_ids):
            ids_str = "**`" + "`**, **`".join(app_ids) + "`**"
            st.error(f"Got invalid App IDs: {ids_str}")
            _reset_app_ids()
            return versions_df

        ids_str = "**`" + "`**, **`".join(app_ids) + "`**"
        info_col, show_all_col = st.columns(
            [0.9, 0.1], vertical_alignment="center"
        )
        info_col.info(f"Filtering with App IDs: {ids_str}")
        versions_df = versions_df[versions_df["app_id"].isin(app_ids)]

        show_all_col.button(
            "Show All", use_container_width=True, on_click=_reset_app_ids
        )
        if not st.query_params.get("app_ids", None):
            st.query_params["app_ids"] = ",".join(app_ids)
    return versions_df


def _handle_record_query_change():
    value = st.session_state.get(f"{page_name}.record_search", None)
    if value:
        st.query_params["record_search"] = value
    else:
        st.query_params.pop("record_search")


def render_records(app_name: str):
    st.title(page_name)
    st.markdown(f"Showing app `{app_name}`")

    # Get app versions
    record_query = st.text_input(
        "Search Records",
        key=f"{page_name}.record_search",
        on_change=_handle_record_query_change,
    )
    versions_df, version_metadata_col_names = render_app_version_filters(
        app_name,
        {f"{page_name}.record_search": record_query},
        page_name_keys=[f"{page_name}.record_search"],
    )

    st.divider()

    versions_df = _render_app_id_args_filter(versions_df)

    if versions_df.empty:
        st.error(f"No app versions found for app `{app_name}`.")
        return
    app_ids = versions_df["app_id"].tolist()

    # Get records and feedback data
    records_limit = st.session_state.get(ST_RECORDS_LIMIT, None)
    records_df, feedback_col_names = get_records_and_feedback(
        app_name=app_name, app_ids=app_ids, limit=records_limit
    )

    feedback_col_names = list(feedback_col_names)

    # Preprocess data
    df = _preprocess_df(
        records_df,
        record_query=record_query,
    )

    if df.empty:
        if record_query:
            st.error(f"No records found for search query `{record_query}`.")
        elif app_ids := st.session_state.get(f"{page_name}.app_ids", None):
            app_versions = versions_df[versions_df["app_id"].isin(app_ids)][
                "app_version"
            ].unique()
            versions_str = "**`" + "`**, **`".join(app_versions) + "`**"
            st.error(f"No records found for app version(s): {versions_str}.")
        else:
            st.error(f"No records found for app `{app_name}`.")
        return
    elif records_limit is not None:
        cols = st.columns([0.9, 0.1], vertical_alignment="center")
        cols[0].info(
            f"Limiting to the latest {records_limit} records. Use the search bar and filters to narrow your search.",
            icon="ℹ️",
        )

        def handle_show_all():
            st.session_state[ST_RECORDS_LIMIT] = None
            if ST_RECORDS_LIMIT in st.query_params:
                del st.query_params[ST_RECORDS_LIMIT]

        cols[1].button(
            "Show all",
            use_container_width=True,
            on_click=handle_show_all,
            help="Show all records. This may take a while.",
        )

    else:
        st.success(f"Found {len(df)} records.")

    _, feedback_directions = get_feedback_defs()

    _render_grid_tab(
        df,
        feedback_col_names,
        feedback_directions,
        version_metadata_col_names,
    )


if __name__ == "__main__":
    set_page_config(page_title=page_name)
    init_page_state()
    app_name = render_sidebar()
    if app_name:
        render_records(app_name)



import argparse
import json
import sys
from typing import Any, Callable, Dict, List, Optional

import pandas as pd
import streamlit as st
from trulens.core import __package__ as core_package
from trulens.core import __version__ as core_version
from trulens.core.database import base as mod_db
from trulens.core.session import TruSession
from trulens.core.utils.imports import static_resource
from trulens.dashboard import __package__ as dashboard_package
from trulens.dashboard import __version__
from trulens.dashboard.constants import CACHE_TTL
from trulens.dashboard.constants import EXTERNAL_APP_COL_NAME
from trulens.dashboard.constants import HIDE_RECORD_COL_NAME
from trulens.dashboard.constants import PINNED_COL_NAME
from trulens.dashboard.constants import RECORDS_LIMIT
from trulens.dashboard.utils.metadata_utils import flatten_metadata

ST_APP_NAME = "app_name"
ST_RECORDS_LIMIT = "records_limit"


def set_page_config(page_title: Optional[str] = None):
    page_title = f"TruLens: {page_title}" if page_title else "TruLens"
    st.set_page_config(
        page_title=page_title,
        page_icon="https://www.trulens.org/img/favicon.ico",
        layout="wide",
    )

    if st.get_option("theme.base") == "dark":
        logo = str(static_resource("dashboard", "ux/trulens_logo_light.svg"))
        logo_small = str(
            static_resource("dashboard", "ux/trulens_squid_light.svg")
        )
    else:
        logo = str(static_resource("dashboard", "ux/trulens_logo.svg"))
        logo_small = str(static_resource("dashboard", "ux/trulens_squid.svg"))

    st.logo(logo, icon_image=logo_small, link="https://www.trulens.org/")

    if ST_RECORDS_LIMIT not in st.session_state:
        st.session_state[ST_RECORDS_LIMIT] = RECORDS_LIMIT


def add_query_param(param_name: str, param_value: str):
    st.query_params[param_name] = param_value


def read_query_params_into_session_state(
    page_name: str, transforms: Optional[dict[str, Callable[[str], Any]]] = None
):
    """This method loads query params into the session state. This function should only be called only once when the page is first initialized.

    Args:
        page_name (str): Name of the page being initialized. Used to prefix page-specific session keys.
        transforms (Optional[dict[str, Callable]], optional): An optional dictionary mapping query param names to a function that deserializes the respective query arg value. Defaults to None.

    """
    assert not st.session_state.get(f"{page_name}.initialized", False)
    for param, value in st.query_params.to_dict().items():
        prefix_page_name = True
        if param == ST_APP_NAME:
            prefix_page_name = False
        elif param.startswith("filter."):
            prefix_page_name = False
            if param.endswith(".multiselect"):
                value = value.split(",")
                if len(value) == 1 and value[0] == "":
                    value = []
        elif transforms and param in transforms:
            value = transforms[param](value)

        if prefix_page_name:
            st.session_state[f"{page_name}.{param}"] = value
        else:
            st.session_state[param] = value


@st.cache_resource(show_spinner="Setting up TruLens session")
def get_session() -> TruSession:
    """Parse command line arguments and initialize TruSession with them.

    As TruSession is a singleton, further TruSession() uses will get the same configuration.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("--database-url", default=None)
    parser.add_argument(
        "--database-prefix", default=mod_db.DEFAULT_DATABASE_PREFIX
    )

    try:
        args = parser.parse_args()
    except SystemExit as e:
        print(e)

        # This exception will be raised if --help or invalid command line arguments
        # are used. Currently, streamlit prevents the program from exiting normally,
        # so we have to do a hard exit.
        sys.exit(e.code)

    return TruSession(
        database_url=args.database_url, database_prefix=args.database_prefix
    )


@st.cache_data(ttl=CACHE_TTL, show_spinner="Getting record data")
def get_records_and_feedback(
    app_name: str,
    app_ids: Optional[List[str]] = None,
    limit: Optional[int] = None,
):
    session = get_session()
    lms = session.connector.db
    assert lms
    records_df, feedback_col_names = lms.get_records_and_feedback(
        app_name=app_name, app_ids=app_ids, limit=limit
    )

    record_json = records_df["record_json"].apply(json.loads)
    records_df["record_metadata"] = record_json.apply(
        lambda x: flatten_metadata(x["meta"])
        if isinstance(x["meta"], dict)
        else {}
    )

    records_df, _ = _factor_out_metadata(records_df, "record_metadata")

    if HIDE_RECORD_COL_NAME in records_df.columns:
        records_df[HIDE_RECORD_COL_NAME] = (
            records_df[HIDE_RECORD_COL_NAME] == "True"
        ).astype(bool)
    records_df = records_df.replace({float("nan"): None})
    return records_df, feedback_col_names


@st.cache_data(ttl=CACHE_TTL, show_spinner="Getting app data")
def get_apps(app_name: Optional[str] = None):
    session = get_session()
    lms = session.connector.db
    assert lms
    return list(lms.get_apps(app_name=app_name))


@st.cache_data(ttl=CACHE_TTL, show_spinner="Getting feedback definitions")
def get_feedback_defs():
    session = get_session()
    lms = session.connector.db
    assert lms

    feedback_defs = lms.get_feedback_defs()
    feedback_directions = {
        (
            row.feedback_json.get("supplied_name", "")
            or row.feedback_json["implementation"]["name"]
        ): row.feedback_json.get("higher_is_better", True)
        for _, row in feedback_defs.iterrows()
    }
    feedback_defs["feedback_name"] = feedback_defs["feedback_json"].map(
        lambda x: x.get("supplied_name", "")
    )
    return feedback_defs, feedback_directions


def update_app_metadata(app_id: str, metadata: dict):
    session = get_session()
    lms = session.connector.db
    assert lms
    lms.update_app_metadata(app_id, metadata)


def _handle_app_selection(app_names: List[str]):
    value = st.session_state.get(f"{ST_APP_NAME}_selector", None)
    if value and value in app_names:
        st.session_state[ST_APP_NAME] = value


def render_refresh_button():
    if st.sidebar.button("↻ Refresh Data", use_container_width=True):
        st.cache_data.clear()
        st.rerun()


def render_sidebar():
    apps = get_apps()
    app_name = st.session_state.get(ST_APP_NAME, None)

    if apps:
        app_names = sorted(
            list(set(app["app_name"] for app in apps)), reverse=True
        )

        if len(app_names) > 1:
            if not app_name or app_name not in app_names:
                app_idx = 0
            else:
                app_idx = app_names.index(app_name)
            if app_name := st.sidebar.selectbox(
                "Select an app",
                index=app_idx,
                key=f"{ST_APP_NAME}_selector",
                options=app_names,
                on_change=_handle_app_selection,
                args=(app_names,),
                disabled=len(app_names) == 1,
            ):
                st.query_params[ST_APP_NAME] = app_name

        else:
            app_name = app_names[0]

    render_refresh_button()

    with st.sidebar.expander("Info"):
        st.text(f"{core_package} {core_version}")
        st.text(f"{dashboard_package} {__version__}")

        st.link_button(
            "Share Feedback",
            "https://forms.gle/HAc4HBk5nZRpgw7C6",
            help="Help us improve TruLens!",
            use_container_width=True,
        )
    if app_name is None:
        st.error("No apps found in the database.")
    return app_name


def _factor_out_metadata(df: pd.DataFrame, metadata_col_name: str):
    metadata_cols = set()
    for _, row in df.iterrows():
        metadata_cols.update(row[metadata_col_name].keys())

    for metadata_key in metadata_cols:
        df[metadata_key] = df[metadata_col_name].apply(
            lambda x: x.get(metadata_key, None)
        )
    return df, metadata_cols


@st.cache_data(ttl=CACHE_TTL, show_spinner="Getting app versions")
def get_app_versions(app_name: str):
    app_versions = get_apps(app_name=app_name)
    app_versions_df = pd.DataFrame(app_versions)

    # Flatten metadata
    app_versions_df["metadata"] = app_versions_df["metadata"].apply(
        lambda x: flatten_metadata(x) if isinstance(x, dict) else {}
    )

    # Factor out metadata
    app_versions_df, app_version_metadata_cols = _factor_out_metadata(
        app_versions_df, "metadata"
    )

    app_versions_df = app_versions_df.replace({float("nan"): None})

    for bool_col in [PINNED_COL_NAME, EXTERNAL_APP_COL_NAME]:
        if bool_col in app_versions_df.columns:
            app_versions_df[bool_col] = (
                app_versions_df[bool_col] == "True"
            ).astype(bool)
    return app_versions_df, list(app_version_metadata_cols)


def _get_query_args_handler(key: str, max_options: Optional[int] = None):
    new_val = st.session_state.get(key)
    if isinstance(new_val, list):
        if len(new_val) == max_options:
            # don't need to explicitly add query args as default is all options
            del st.query_params[key]
            return
        new_val = ",".join(str(v) for v in new_val)
    elif not isinstance(new_val, str):
        raise ValueError(
            f"Unable to save value to query params: {new_val} (type: {type(new_val)})"
        )
    st.query_params[key] = new_val


def _render_filter_multiselect(name: str, options: List[str], key: str):
    return st.multiselect(
        name,
        options,
        default=options,
        key=key,
        on_change=_get_query_args_handler,
        args=(
            key,
            len(options),
        ),
    )


def _handle_reset_filters(
    keys: List[str],
    tags: List[str],
    metadata_options: Dict[str, List[str]],
    page_name_keys: Optional[List[str]] = None,
):
    for key in keys:
        if key == "filter.tags.multiselect":
            val = tags
        elif key.startswith("filter.metadata.") and key.endswith(
            ".multiselect"
        ):
            metadata_key = key[16:-12]
            val = metadata_options[metadata_key]
        else:
            val = ""

        st.session_state[key] = val
        query_param_key = key
        if page_name_keys and key in page_name_keys:
            query_param_key = ".".join(query_param_key.split(".")[1:])
        del st.query_params[query_param_key]


def render_app_version_filters(
    app_name: str,
    other_query_params_kv: Optional[dict[str, str]] = None,
    page_name_keys: Optional[List[str]] = None,
):
    app_versions_df, app_version_metadata_cols = get_app_versions(app_name)
    filtered_app_versions = app_versions_df

    col0, col1, col2 = st.columns(
        [0.7, 0.15, 0.15], vertical_alignment="bottom"
    )
    if other_query_params_kv:
        active_adv_filters = [k for k, v in other_query_params_kv.items() if v]
    else:
        active_adv_filters = []
    if version_str_query := col0.text_input(
        "Search App Version",
        key="filter.search",
        on_change=_get_query_args_handler,
        args=("filter.search",),
    ):
        active_adv_filters.append("filter.search")
        filtered_app_versions = filtered_app_versions[
            filtered_app_versions["app_version"].str.contains(
                version_str_query, case=False
            )
        ]

    with col1.popover("Advanced Filters", use_container_width=True):
        # get tag options
        st.header("Advanced Filters")
        tags = set()
        for _, app_version in app_versions_df.iterrows():
            tags |= set(app_version["tags"])
        tags = sorted(tags)
        # select tags

        selected_tags = _render_filter_multiselect(
            "tags", tags, key="filter.tags.multiselect"
        )
        if len(selected_tags) != len(tags):
            active_adv_filters.append("filter.tags.multiselect")

        metadata_options = {}
        for metadata_key in app_version_metadata_cols:
            try:
                unique_values = app_versions_df[metadata_key].unique()
            except TypeError:
                pass
            if len(unique_values):
                metadata_options[metadata_key] = list(unique_values)

        # select metadata
        metadata_selections = {}
        for metadata_key in metadata_options.keys():
            _metadata_select_options = sorted(
                metadata_options[metadata_key], key=str
            )
            metadata_selections[metadata_key] = _render_filter_multiselect(
                metadata_key,
                _metadata_select_options,
                key=f"filter.metadata.{metadata_key}.multiselect",
            )
            if len(metadata_selections[metadata_key]) != len(
                _metadata_select_options
            ):
                active_adv_filters.append(
                    f"filter.metadata.{metadata_key}.multiselect"
                )

        # filter to apps with selected metadata
        for metadata_key in metadata_selections.keys():
            filtered_app_versions = filtered_app_versions[
                filtered_app_versions[metadata_key].isin(
                    metadata_selections[metadata_key]
                )
            ]
        filtered_app_versions = filtered_app_versions[
            filtered_app_versions["tags"].apply(
                lambda x: any(tag in x for tag in selected_tags)
            )
        ]

    if len(active_adv_filters):
        col2.button(
            "Reset Filters",
            use_container_width=True,
            type="primary",
            on_click=_handle_reset_filters,
            args=(active_adv_filters, tags, metadata_options, page_name_keys),
        )

    return filtered_app_versions, app_version_metadata_cols



import math
import re
from typing import Any, Dict, List, Optional, Sequence

import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from st_aggrid import AgGrid
from st_aggrid.grid_options_builder import GridOptionsBuilder
import streamlit as st
from trulens.apps.virtual import TruVirtual
from trulens.apps.virtual import VirtualApp
from trulens.apps.virtual import VirtualRecord
from trulens.core.schema.feedback import FeedbackResult
from trulens.core.schema.feedback import FeedbackResultStatus
from trulens.core.utils.text import format_quantity
from trulens.dashboard.constants import COMPARE_PAGE_NAME as compare_page_name
from trulens.dashboard.constants import EXTERNAL_APP_COL_NAME
from trulens.dashboard.constants import HIDE_RECORD_COL_NAME
from trulens.dashboard.constants import LEADERBOARD_PAGE_NAME as page_name
from trulens.dashboard.constants import PINNED_COL_NAME
from trulens.dashboard.constants import RECORDS_PAGE_NAME as records_page_name
from trulens.dashboard.pages.Compare import MAX_COMPARATORS
from trulens.dashboard.pages.Compare import MIN_COMPARATORS
from trulens.dashboard.utils.dashboard_utils import ST_RECORDS_LIMIT
from trulens.dashboard.utils.dashboard_utils import get_app_versions
from trulens.dashboard.utils.dashboard_utils import get_apps
from trulens.dashboard.utils.dashboard_utils import get_feedback_defs
from trulens.dashboard.utils.dashboard_utils import get_records_and_feedback
from trulens.dashboard.utils.dashboard_utils import get_session
from trulens.dashboard.utils.dashboard_utils import (
    read_query_params_into_session_state,
)
from trulens.dashboard.utils.dashboard_utils import render_app_version_filters
from trulens.dashboard.utils.dashboard_utils import render_sidebar
from trulens.dashboard.utils.dashboard_utils import set_page_config
from trulens.dashboard.utils.dashboard_utils import update_app_metadata
from trulens.dashboard.utils.metadata_utils import nest_dict
from trulens.dashboard.utils.metadata_utils import nest_metadata
from trulens.dashboard.utils.metadata_utils import nested_update
from trulens.dashboard.ux.components import draw_metadata_and_tags
from trulens.dashboard.ux.styles import CATEGORY
from trulens.dashboard.ux.styles import Category
from trulens.dashboard.ux.styles import aggrid_css
from trulens.dashboard.ux.styles import cell_rules
from trulens.dashboard.ux.styles import default_direction
from trulens.dashboard.ux.styles import stmetricdelta_hidearrow

APP_COLS = ["app_version", "app_id", "app_name"]
APP_AGG_COLS = ["Records", "Average Latency"]


def init_page_state():
    if st.session_state.get(f"{page_name}.initialized", False):
        return

    read_query_params_into_session_state(
        page_name=page_name,
        transforms={
            "metadata_to_front": lambda x: x == "True",
            "only_show_pinned": lambda x: x == "True",
            "metadata_cols": lambda x: x.split(","),
        },
    )
    st.session_state[f"{page_name}.initialized"] = True


def _preprocess_df(
    records_df: pd.DataFrame,
    app_versions_df: pd.DataFrame,
    feedback_col_names: List[str],
    metadata_col_names: List[str],
    show_all: bool = False,
):
    records_df = records_df.sort_values(by="app_id")

    records_df["total_cost_usd"] = records_df["total_cost"].where(
        records_df["cost_currency"] == "USD",
        other=0,
    )
    records_df["total_cost_sf"] = records_df["total_cost"].where(
        records_df["cost_currency"] == "Snowflake credits",
        other=0,
    )

    agg_dict = {
        "Records": ("record_id", "count"),
        "Average Latency": ("latency", "mean"),
        "Total Cost (USD)": ("total_cost_usd", "sum"),
        "Total Cost (Snowflake Credits)": ("total_cost_sf", "sum"),
        "Total Tokens": ("total_tokens", "sum"),
        "tags": ("tags", lambda x: ",".join(x.drop_duplicates())),
    }
    for col in feedback_col_names:
        agg_dict[col] = (col, "mean")

    app_agg_df: pd.DataFrame = (
        records_df.groupby(
            by=["app_version", "app_name", "app_id"], dropna=True, sort=True
        )
        .aggregate(**agg_dict)
        .reset_index()
    )

    if PINNED_COL_NAME in app_versions_df:
        app_versions_df[PINNED_COL_NAME] = app_versions_df[PINNED_COL_NAME]

    df = app_agg_df.join(
        app_versions_df.set_index(["app_id", "app_name", "app_version"])[
            metadata_col_names
        ],
        how="left" if not show_all else "right",
        validate="many_to_one",
        on=["app_id", "app_name", "app_version"],
    ).round(3)
    return df


def order_columns(
    df: pd.DataFrame,
    order: Sequence[str],
):
    return df[order]


def _build_grid_options(
    df: pd.DataFrame,
    feedback_col_names: List[str],
    feedback_directions: Dict[str, bool],
    version_metadata_col_names: Sequence[str],
):
    gb = GridOptionsBuilder.from_dataframe(df, headerHeight=50)

    gb.configure_column(
        "app_version",
        header_name="App Version",
        resizable=True,
        pinned="left",
        filter="agMultiColumnFilter",
    )

    gb.configure_columns(APP_COLS, filter="agMultiColumnFilter")
    gb.configure_columns(
        feedback_col_names + ["records", "latency"],
        filter="agNumberColumnFilter",
    )
    gb.configure_columns(
        version_metadata_col_names, filter="agMultiColumnFilter", editable=True
    )

    gb.configure_column(
        PINNED_COL_NAME,
        header_name="Pinned",
        hide=True,
        filter="agSetColumnFilter",
    )
    gb.configure_column(
        EXTERNAL_APP_COL_NAME,
        header_name="External App",
        hide=True,
        filter="agSetColumnFilter",
    )
    gb.configure_column(
        "app_id",
        header_name="App ID",
        hide=True,
        resizable=True,
        filter="agSetColumnFilter",
    )
    gb.configure_column(
        "app_name",
        header_name="App Name",
        hide=True,
        resizable=True,
        filter="agMultiColumnFilter",
    )

    for feedback_col in feedback_col_names:
        if "distance" in feedback_col:
            gb.configure_column(
                feedback_col,
                hide=feedback_col.endswith("_calls"),
                filter="agNumberColumnFilter",
            )
        else:
            # cell highlight depending on feedback direction
            feedback_direction = (
                "HIGHER_IS_BETTER"
                if feedback_directions.get(feedback_col, default_direction)
                else "LOWER_IS_BETTER"
            )

            gb.configure_column(
                feedback_col,
                cellClassRules=cell_rules[feedback_direction],
                hide=feedback_col.endswith("_calls"),
                filter="agNumberColumnFilter",
            )

    gb.configure_grid_options(
        rowHeight=45,
        suppressContextMenu=True,
        rowClassRules={
            # "external-app": f"data['{EXTERNAL_APP_COL_NAME}'] > 0",
            "app-external": f"data['{EXTERNAL_APP_COL_NAME}']",
            "app-pinned": f"data['{PINNED_COL_NAME}']",
        },
    )
    gb.configure_selection(
        selection_mode="multiple",
        use_checkbox=True,
    )
    gb.configure_pagination(enabled=True, paginationPageSize=25)
    gb.configure_side_bar()
    gb.configure_grid_options(
        autoSizeStrategy={"type": "fitCellContents", "skipHeader": False}
    )
    return gb.build()


def _render_grid(
    df: pd.DataFrame,
    feedback_col_names: List[str],
    feedback_directions: Dict[str, bool],
    version_metadata_col_names: Sequence[str],
    grid_key: Optional[str] = None,
):
    columns_state = st.session_state.get(f"{grid_key}.columns_state", None)

    if PINNED_COL_NAME in df:
        df.loc[df[PINNED_COL_NAME], "app_version"] = df.loc[
            df[PINNED_COL_NAME], "app_version"
        ].apply(lambda x: f"📌 {x}")

    height = 1000 if len(df) > 20 else 45 * len(df) + 100

    return AgGrid(
        df,
        key=grid_key,
        height=height,
        columns_state=columns_state,
        gridOptions=_build_grid_options(
            df=df,
            feedback_col_names=feedback_col_names,
            feedback_directions=feedback_directions,
            version_metadata_col_names=version_metadata_col_names,
        ),
        custom_css=aggrid_css,
        update_on=["selectionChanged", "cellValueChanged"],
        allow_unsafe_jscode=True,
    )


def handle_pin_toggle(selected_app_ids: List[str], on_leaderboard: bool):
    # Create nested metadata dict
    value = nest_metadata(PINNED_COL_NAME, not on_leaderboard)
    for app_id in selected_app_ids:
        update_app_metadata(app_id, value)
    get_app_versions.clear()
    get_apps.clear()
    if on_leaderboard:
        st.toast(
            f"Successfully removed {len(selected_app_ids)} app(s) from Leaderboard"
        )
    else:
        st.toast(
            f"Successfully added {len(selected_app_ids)} app(s) to Leaderboard"
        )


def handle_table_edit(
    df: pd.DataFrame,
    event_data: Dict[str, Any],
    version_metadata_col_names: List[str],
):
    app_id = event_data["data"]["app_id"]
    if app_id not in df["app_id"].values:
        st.error(f"App with ID {app_id} not found in the leaderboard!")
        return

    app_df = df[df["app_id"] == app_id].iloc[0]
    metadata = {}
    for col in version_metadata_col_names:
        if col in event_data["data"] and event_data["data"][col] != app_df[col]:
            value = nest_metadata(col, event_data["data"][col])
            nested_update(metadata, value)
    update_app_metadata(app_id, metadata)

    get_app_versions.clear()
    get_apps.clear()
    st.toast(f"Successfully updated metadata for `{app_df['app_version']}`")


@st.dialog("Add/Edit Metadata")
def handle_add_metadata(
    selected_rows: pd.DataFrame, metadata_col_names: List[str]
):
    st.write(
        f"Add or edit metadata for {len(selected_rows)} selected app version(s)"
    )
    key = st.text_input("Metadata Key", placeholder="metadata.key")

    if key and not re.match(r"^[A-Za-z0-9_.]+$", key):
        st.error(
            "`key` must contain only alphanumeric and underscore characters!"
        )

    existing_value = None
    placeholder = None
    if key in selected_rows.columns:
        existing_values = list(selected_rows[key].unique())
        if len(existing_values) == 1:
            existing_value = existing_values[0]
        else:
            placeholder = "<multiple existing values>"

    val = st.text_input(
        "Metadata Value",
        placeholder=placeholder or "value",
        value=existing_value,
    )

    if st.button("Submit"):
        if not key:
            st.error("Metadata key cannot be empty!")
            return
        metadata = nest_metadata(key, val)
        for app_id in selected_rows["app_id"]:
            update_app_metadata(app_id, metadata)

        if key not in metadata_col_names:
            metadata_col_names.append(key)

        get_app_versions.clear()
        get_apps.clear()
        st.toast(
            f"Successfully updated metadata for {len(selected_rows)} app(s)"
        )
        st.rerun()


@st.dialog("Add Virtual App")
def handle_add_virtual_app(
    app_name: str,
    feedback_col_names: List[str],
    feedback_defs: Any,
    metadata_col_names: List[str],
):
    with st.form(f"{page_name}.add_virtual_app_form", border=False):
        app_version = st.text_input(
            "App Version", placeholder="virtual_app_base"
        )

        feedback_values = {
            col: st.number_input(col)
            for col in feedback_col_names
            if col in feedback_defs["feedback_name"].values
        }
        metadata_values = nest_dict({
            col: st.text_input(col)
            for col in metadata_col_names
            if not col.startswith("trulens.")
        })

        metadata_values = {k: v for k, v in metadata_values.items() if v}
        metadata_values[EXTERNAL_APP_COL_NAME] = True

        if st.form_submit_button("Submit"):
            if app_version and not re.match(r"^[A-Za-z0-9_.]+$", app_version):
                st.error(
                    "`app_version` must contain only alphanumeric and underscore characters!"
                )
                return
            session = get_session()
            app = TruVirtual(
                app=VirtualApp(),
                app_name=app_name,
                app_version=app_version,
                metadata=metadata_values,
            )

            virtual_record = VirtualRecord(
                calls={},
                main_input="<autogenerated record>",
                main_output="<autogenerated record>",
                meta=nest_metadata(HIDE_RECORD_COL_NAME, True),
            )

            app.add_record(virtual_record)
            for feedback_name, feedback_value in feedback_values.items():
                result = FeedbackResult(
                    record_id=virtual_record.record_id,
                    feedback_definition_id=feedback_defs[
                        feedback_defs["feedback_name"] == feedback_name
                    ]["feedback_definition_id"].iloc[0],
                    status=FeedbackResultStatus.DONE,
                    cost={},
                    perf={},
                    name=feedback_name,
                    result=feedback_value,
                )
                session.connector.db.insert_feedback(result)

            get_records_and_feedback.clear()
            get_app_versions.clear()
            get_apps.clear()
            st.toast(f"Successfully created virtual app version {app_version}.")
            st.rerun()


def _render_grid_tab(
    df: pd.DataFrame,
    feedback_col_names: List[str],
    feedback_defs: Any,
    feedback_directions: Dict[str, bool],
    version_metadata_col_names: List[str],
    app_name: str,
    grid_key: Optional[str] = None,
):
    container = st.container()
    c1, c2, c3, c4, c5, c6 = container.columns(
        [1, 1, 1, 1, 1, 1],
        vertical_alignment="center",
    )

    _metadata_options = [
        col
        for col in version_metadata_col_names
        if not col.startswith("trulens.")
    ]

    # Validate metadata_cols
    if metadata_cols := st.session_state.get(f"{page_name}.metadata_cols", []):
        st.session_state[f"{page_name}.metadata_cols"] = [
            col_name
            for col_name in metadata_cols
            if col_name in _metadata_options
        ]

    metadata_cols = st.multiselect(
        label="Display Metadata Columns",
        key=f"{page_name}.metadata_cols",
        options=_metadata_options,
        default=_metadata_options,
    )
    if len(metadata_cols) != len(_metadata_options):
        st.query_params["metadata_cols"] = ",".join(metadata_cols)
    if (
        EXTERNAL_APP_COL_NAME in df
        and EXTERNAL_APP_COL_NAME not in metadata_cols
    ):
        metadata_cols.append(EXTERNAL_APP_COL_NAME)
    if PINNED_COL_NAME in df and PINNED_COL_NAME not in metadata_cols:
        metadata_cols.append(PINNED_COL_NAME)

    if metadata_to_front := c1.toggle(
        "Metadata to Front",
        key=f"{page_name}.metadata_to_front",
    ):
        df = order_columns(
            df,
            APP_COLS + metadata_cols + APP_AGG_COLS + feedback_col_names,
        )
    else:
        df = order_columns(
            df,
            APP_COLS + APP_AGG_COLS + feedback_col_names + metadata_cols,
        )
    st.query_params["metadata_to_front"] = str(metadata_to_front)

    if only_show_pinned := c1.toggle(
        "Only Show Pinned",
        key=f"{page_name}.only_show_pinned",
    ):
        if PINNED_COL_NAME in df:
            df = df[df[PINNED_COL_NAME]]
        else:
            st.info(
                "Pin an app version by selecting it and clicking the `Pin App` button.",
                icon="📌",
            )
            return
    st.query_params["only_show_pinned"] = str(only_show_pinned)

    grid_data = _render_grid(
        df,
        feedback_col_names=feedback_col_names,
        feedback_directions=feedback_directions,
        version_metadata_col_names=version_metadata_col_names,
        grid_key=grid_key,
    )

    if (
        grid_data.event_data
        and grid_data.event_data["type"] == "cellValueChanged"
    ):
        handle_table_edit(df, grid_data.event_data, version_metadata_col_names)

    selected_rows = grid_data.selected_rows
    selected_rows = pd.DataFrame(selected_rows)
    if selected_rows.empty:
        selected_app_ids = []
    else:
        selected_app_ids = list(selected_rows.app_id.unique())

    # Add to Leaderboard
    on_leaderboard = any(
        PINNED_COL_NAME in app and app[PINNED_COL_NAME]
        for _, app in selected_rows.iterrows()
    )
    if c2.button(
        "Unpin App" if on_leaderboard else "Pin App",
        key=f"{page_name}.pin_button",
        disabled=selected_rows.empty,
        on_click=handle_pin_toggle,
        use_container_width=True,
        args=(selected_app_ids, on_leaderboard),
    ):
        st.rerun()
    # Examine Records
    if c3.button(
        "Examine Records",
        disabled=selected_rows.empty,
        use_container_width=True,
        key=f"{page_name}.records_button",
    ):
        st.session_state[f"{records_page_name}.app_ids"] = selected_app_ids
        st.switch_page("pages/Records.py")
    # Compare App Versions
    if len(selected_app_ids) < MIN_COMPARATORS:
        _compare_button_label = f"Min {MIN_COMPARATORS} App Versions"
        _compare_button_disabled = True
        help_msg = f"Select at least {MIN_COMPARATORS} app versions to compare."
    elif len(selected_app_ids) > MAX_COMPARATORS:
        _compare_button_label = f"Max {MAX_COMPARATORS} App Versions"
        _compare_button_disabled = True
        help_msg = (
            f"Deselect to at most {MAX_COMPARATORS} app versions to compare."
        )
    else:
        _compare_button_label = "Compare"
        _compare_button_disabled = False
        help_msg = None

    if c4.button(
        _compare_button_label,
        help=help_msg,
        disabled=_compare_button_disabled,
        use_container_width=True,
        key=f"{page_name}.compare_button",
    ):
        st.session_state[f"{compare_page_name}.app_ids"] = selected_app_ids
        st.switch_page("pages/Compare.py")

    # Add Metadata Col
    if c5.button(
        "Add/Edit Metadata",
        disabled=selected_rows.empty,
        use_container_width=True,
        key=f"{page_name}.modify_metadata_button",
    ):
        handle_add_metadata(selected_rows, version_metadata_col_names)

    # Add Virtual App
    if c6.button(
        "Add Virtual App",
        use_container_width=True,
        key=f"{page_name}.add_virtual_app_button",
    ):
        handle_add_virtual_app(
            app_name,
            feedback_col_names,
            feedback_defs,
            version_metadata_col_names,
        )


@st.fragment
def _render_list_tab(
    df: pd.DataFrame,
    feedback_col_names: List[str],
    feedback_directions: Dict[str, bool],
    version_metadata_col_names: List[str],
    max_feedback_cols: int = 6,
):
    st.markdown(
        stmetricdelta_hidearrow,
        unsafe_allow_html=True,
    )
    for _, app_row in df.iterrows():
        app_id = app_row["app_id"]
        app_version = app_row["app_version"]
        tags = app_row["tags"]
        metadata = {
            col: app_row[col]
            for col in version_metadata_col_names
            if col in app_row
        }
        st.markdown(
            f"#### {app_version}", help=draw_metadata_and_tags(metadata, tags)
        )
        st.caption(app_id)
        app_feedback_col_names = [
            col_name
            for col_name in feedback_col_names
            if col_name in app_row and app_row[col_name] is not None
        ]
        (
            n_records_col,
            latency_col,
            tokens_col,
            cost_col,
            select_app_col,
        ) = st.columns([1, 1, 1, 1, 1])
        feedback_cols = st.columns(
            min(len(app_feedback_col_names), max_feedback_cols)
        )

        n_records_col.metric("Records", app_row["Records"])

        latency_mean = app_row["Average Latency"]
        latency_col.metric(
            "Average Latency (Seconds)",
            (
                f"{format_quantity(round(latency_mean, 5), precision=2)}"
                if not math.isnan(latency_mean)
                else "nan"
            ),
        )

        if app_row["Total Cost (Snowflake Credits)"] > 0:
            cost_col.metric(
                "Total Cost (Snowflake credits)",
                f"{format_quantity(round(app_row['Total Cost (Snowflake Credits)'], 8), precision=5)}",
            )
        elif app_row["Total Cost (USD)"] > 0:
            cost_col.metric(
                "Total Cost (USD)",
                f"${format_quantity(round(app_row['Total Cost (USD)'], 5), precision=2)}",
            )

        tokens_col.metric(
            "Total Tokens",
            format_quantity(
                app_row["Total Tokens"],
                precision=2,
            ),
        )

        col_counter = 0
        for col_name in app_feedback_col_names:
            mean = app_row[col_name]
            if mean is None or pd.isna(mean):
                continue
            col = feedback_cols[col_counter % max_feedback_cols]
            col_counter += 1
            feedback_container = col.container(border=True)

            higher_is_better = feedback_directions.get(col_name, True)

            if "distance" in col_name:
                feedback_container.metric(
                    label=col_name,
                    value=f"{round(mean, 2)}",
                    delta_color="normal",
                )
            else:
                cat: Category = CATEGORY.of_score(
                    mean, higher_is_better=higher_is_better
                )
                feedback_container.metric(
                    label=col_name,
                    value=f"{round(mean, 2)}",
                    delta=f"{cat.icon} {cat.adjective}",
                    delta_color=(
                        "normal"
                        if cat.compare
                        and cat.direction
                        and cat.compare(
                            mean, CATEGORY.PASS[cat.direction].threshold
                        )
                        else "inverse"
                    ),
                )

        with select_app_col:
            if st.button(
                "Select App",
                key=f"select_app_{app_id}",
            ):
                st.session_state[f"{records_page_name}.app_ids"] = [app_id]
                st.switch_page("pages/Records.py")

        # with st.expander("Model metadata"):
        #    st.markdown(draw_metadata(metadata))

        st.markdown("""---""")


@st.fragment
def _render_plot_tab(df: pd.DataFrame, feedback_col_names: List[str]):
    if HIDE_RECORD_COL_NAME in df.columns:
        df = df[~df[HIDE_RECORD_COL_NAME]]
    cols = 4
    rows = len(feedback_col_names) // cols + 1
    fig = make_subplots(rows=rows, cols=cols, subplot_titles=feedback_col_names)

    for i, feedback_col_name in enumerate(feedback_col_names):
        row_num = i // cols + 1
        col_num = i % cols + 1
        _df = df[feedback_col_name].dropna()

        plot = go.Histogram(
            x=_df,
            xbins={
                "size": 0.1,
                "start": 0,
                "end": 1.0,
            },
            histfunc="count",
            texttemplate="%{y}",
        )
        fig.add_trace(
            plot,
            row=row_num,
            col=col_num,
        )
        if i == 0:
            xaxis = fig["layout"]["xaxis"]
            yaxis = fig["layout"]["yaxis"]
        else:
            xaxis = fig["layout"][f"xaxis{i + 1}"]
            yaxis = fig["layout"][f"yaxis{i + 1}"]

        xaxis["title"] = "Score"
        if col_num == 1:
            yaxis["title"] = "# Records"

    fig.update_layout(
        height=300 * rows,
        width=200 * cols,
        dragmode=False,
        showlegend=False,
        barcornerradius=4,
        bargap=0.05,
    )
    fig.update_yaxes(fixedrange=True, showgrid=False)
    fig.update_xaxes(fixedrange=True, showgrid=False, range=[0, 1])
    st.plotly_chart(fig, use_container_width=True)


def render_leaderboard(app_name: str):
    st.title(page_name)
    st.markdown(f"Showing app `{app_name}`")

    # Get app versions
    versions_df, version_metadata_col_names = render_app_version_filters(
        app_name
    )
    st.divider()

    if versions_df.empty:
        st.error(f"No app versions found for app `{app_name}`.")
        return
    app_ids = versions_df["app_id"].tolist()

    # Get records and feedback data
    records_limit = st.session_state.get(ST_RECORDS_LIMIT, None)
    records_df, feedback_col_names = get_records_and_feedback(
        app_name=app_name, app_ids=app_ids, limit=records_limit
    )
    if records_df.empty:
        st.error(f"No records found for app `{app_name}`.")
        return
    elif records_limit is not None:
        cols = st.columns([0.9, 0.1], vertical_alignment="center")
        cols[0].info(
            f"Computed from the last {records_limit} records.",
            icon="ℹ️",
        )

        def handle_show_all():
            st.session_state[ST_RECORDS_LIMIT] = None
            if ST_RECORDS_LIMIT in st.query_params:
                del st.query_params[ST_RECORDS_LIMIT]

        cols[1].button(
            "Show all",
            use_container_width=True,
            on_click=handle_show_all,
            help="Show all records. This may take a while.",
        )

    feedback_col_names = list(feedback_col_names)
    # Preprocess data
    df = _preprocess_df(
        records_df,
        versions_df,
        list(feedback_col_names),
        version_metadata_col_names,
    )
    feedback_defs, feedback_directions = get_feedback_defs()

    (
        versions_tab,
        plot_tab,
        list_tab,
    ) = st.tabs([
        "App Versions",
        "Feedback Histograms",
        "List View",
    ])
    with versions_tab:
        _render_grid_tab(
            df,
            grid_key=f"{page_name}.leaderboard_grid",
            feedback_col_names=feedback_col_names,
            feedback_defs=feedback_defs,
            feedback_directions=feedback_directions,
            app_name=app_name,
            version_metadata_col_names=version_metadata_col_names,
        )
    with plot_tab:
        _render_plot_tab(records_df, feedback_col_names)
    with list_tab:
        _render_list_tab(
            df,
            feedback_col_names,
            feedback_directions,
            version_metadata_col_names,
        )


if __name__ == "__main__":
    set_page_config(page_title=page_name)
    init_page_state()
    app_name = render_sidebar()
    if app_name:
        render_leaderboard(app_name)
